"""AsyncJOptional module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Final, Optional, cast, Awaitable
from unittest.mock import patch, MagicMock

from pytest import raises

from dev4py.utils import AsyncJOptional, JOptional
from dev4py.utils.awaitables import to_awaitable, async_none
from dev4py.utils.types import Supplier, Function, Consumer, Runnable, Predicate


class TestAsyncJOptional:
    """AsyncJOptional class tests"""
    NO_VALUE_ERROR_MSG: Final[str] = "No value present"
    CONSTRUCTOR_ERROR_MSG: Final[str] = \
        "AsyncJOptional private constructor! Please use AsyncJOptional.of or AsyncJOptional.of_noneable"

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:

            def test_none_create_key__should__raise_assertion_error(self):
                """When create key is none, should raise AssertionError with private constructor message"""
                # GIVEN / WHEN / THEN
                with raises(AssertionError) as error:
                    AsyncJOptional("A value", None)

                assert str(error.value) == TestAsyncJOptional.CONSTRUCTOR_ERROR_MSG

            def test_invalid_create_key__should__raise_assertion_error(self):
                """When create key is invalid, should raise AssertionError with private constructor message"""
                # GIVEN / WHEN / THEN
                with raises(AssertionError) as error:
                    AsyncJOptional("A value", object())

                assert str(error.value) == TestAsyncJOptional.CONSTRUCTOR_ERROR_MSG

            async def test_valid_create_key__should__create_a_new_async_joptional(self):
                """When create key is valid, should create a new AsyncJOptional"""
                # GIVEN
                value: int = 1

                # WHEN
                # noinspection PyUnresolvedReferences
                optional: AsyncJOptional[int] = AsyncJOptional(value, AsyncJOptional._AsyncJOptional__CREATE_KEY)

                # THEN
                assert await optional.is_present()
                assert await optional.get() == value

    class TestOf:
        """of class method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_async_joptional_with_value(self) -> None:
                """When a value is provided, should return an async joptional with value"""
                # GIVEN
                value: int = 1

                # WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)

                # THEN
                assert await optional.is_present()
                assert await optional.get() == value

            async def test_async_value_exists__should__return_async_joptional_with_value(self) -> None:
                """When an async value is provided, should return an async joptional with value"""
                # GIVEN
                value: int = 1

                # WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(value))

                # THEN
                assert await optional.is_present()
                assert await optional.get() == value

        class TestErrorCase:
            async def test_none_value__should__raise_type_error_on_terminal_operation(self) -> None:
                """When no value is provided, should raise a TypeError exception on terminal operation"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(None)

                # WHEN / THEN
                with raises(TypeError) as error:
                    await optional.get()

                assert str(error.value) == TestAsyncJOptional.NO_VALUE_ERROR_MSG

            async def test_async_none_value__should__raise_type_error_on_terminal_operation(self) -> None:
                """When async none value is provided, should raise a TypeError exception on terminal operation"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_none())

                # WHEN / THEN
                with raises(TypeError) as error:
                    await optional.get()

                assert str(error.value) == TestAsyncJOptional.NO_VALUE_ERROR_MSG

    class TestOfNoneable:
        """of_noneable class method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_async_joptional_with_value(self) -> None:
                """When a value is provided, should return an async joptional with value"""
                # GIVEN
                value: int = 1

                # WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # THEN
                assert await optional.is_present()
                assert await optional.get() == value

            async def test_async_value_exists__should__return_async_joptional_with_value(self) -> None:
                """When an async value is provided, should return an async joptional with value"""
                # GIVEN
                value: int = 1

                # WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(to_awaitable(value))

                # THEN
                assert await optional.is_present()
                assert await optional.get() == value

            async def test_none_value__should__return_empty_async_joptional(self) -> None:
                """When no value is provided, should return an empty async joptional"""
                # GIVEN
                value: Optional[int] = None

                # WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # THEN
                assert optional._value == value
                assert await optional.is_empty()

            async def test_async_none_value__should__return_empty_async_joptional(self) -> None:
                """When async None value is provided, should return an empty async joptional"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # THEN
                assert await optional.is_empty()
                assert await optional.or_else_get() is None

    class TestEmpty:
        """empty class method tests"""

        class TestNominalCase:
            async def test_should__return_empty_async_joptional(self) -> None:
                """Should return an async joptional with value"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # THEN
                assert optional._value is None
                assert await optional.is_empty()

    class TestIsPresent:
        """is_present method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_true(self) -> None:
                """When a value is provided, should return true"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # THEN
                assert await optional.is_present()

            async def test_async_value_exists__should__return_true(self) -> None:
                """When an async value is provided, should return true"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))

                # THEN
                assert await optional.is_present()

            async def test_none_value__should__return_false(self) -> None:
                """When no value is provided, should return false"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # THEN
                assert not await optional.is_present()

            async def test_async_none_value__should__return_false(self) -> None:
                """When async None value is provided, should return false"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # THEN
                assert not await optional.is_present()

    class TestIsEmpty:
        """is_empty method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_false(self) -> None:
                """When a value is provided, should return false"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # THEN
                assert not await optional.is_empty()

            async def test_async_value_exists__should__return_false(self) -> None:
                """When an async value is provided, should return false"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))

                # THEN
                assert not await optional.is_empty()

            async def test_none_value__should__return_true(self) -> None:
                """When no value is provided, should return true"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # THEN
                assert await optional.is_empty()

            async def test_async_none_value__should__return_true(self) -> None:
                """When async None value is provided, should return true"""
                # GIVEN / WHEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # THEN
                assert await optional.is_empty()

    class TestGet:
        """get method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN
                result: int = await optional.get()

                # THEN
                assert result == value

            async def test_async_value_exists__should__return_value(self) -> None:
                """When async value is provided, return the value"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(to_awaitable(value))

                # WHEN
                result: int = await optional.get()

                # THEN
                assert result == value

        class TestErrorCase:
            async def test_none_value__should__raise_type_error(self) -> None:
                """When no value is provided, should raise a ValueError exception with no value message"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(ValueError) as error:
                    await optional.get()

                assert str(error.value) == TestAsyncJOptional.NO_VALUE_ERROR_MSG

            async def test_async_none_value__should__raise_type_error(self) -> None:
                """When async None value is provided, should raise a ValueError exception with no value message"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN / THEN
                with raises(ValueError) as error:
                    await optional.get()

                assert str(error.value) == TestAsyncJOptional.NO_VALUE_ERROR_MSG

    class TestOrElse:
        """or_else method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                default_value: int = 2
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN
                result: int = await optional.or_else(default_value)

                # THEN
                assert result == value

            async def test_async_value_exists__should__return_value(self) -> None:
                """When async value is provided, return the value"""
                # GIVEN
                value: int = 1
                default_value: int = 2
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(to_awaitable(value))

                # WHEN
                result: int = await optional.or_else(default_value)

                # THEN
                assert result == value

            async def test_none_value__should__return_default_value(self) -> None:
                """When no value is provided, return the default value"""
                # GIVEN
                default_value: int = 2
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN
                result: int = await optional.or_else(default_value)

                # THEN
                assert result == default_value

            async def test_async_none_value__should__return_default_value(self) -> None:
                """When async None value is provided, return the default value"""
                # GIVEN
                default_value: int = 2
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                result: int = await optional.or_else(default_value)

                # THEN
                assert result == default_value

            async def test_none_value_and_none_default__should__return_none(self) -> None:
                """When no value is provided and default value is None, return the default value"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN
                result: Optional[int] = await optional.or_else(None)

                # THEN
                assert result is None

            async def test_async_none_value_and_none_default__should__return_none(self) -> None:
                """When async None value is provided and default value is None, return the default value"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                result: Optional[int] = await optional.or_else(None)

                # THEN
                assert result is None

            async def test_none_value_and_no_default__should__return_none(self) -> None:
                """When no value is provided and no default value, return None value"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN
                result: Optional[int] = await optional.or_else()

                # THEN
                assert result is None

            async def test_async_none_value_and_no_default__should__return_none(self) -> None:
                """When async None value is provided and no default value, return None value"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                result: Optional[int] = await optional.or_else()

                # THEN
                assert result is None

    class TestOrElseGet:
        """or_else_get method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                supplier: Supplier[int] = lambda: 2
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN
                result: int = await optional.or_else_get(supplier)

                # THEN
                assert result == value

            async def test_async_value_exists__should__return_value(self) -> None:
                """When async value is provided, return the value"""
                # GIVEN
                value: int = 1
                supplier: Supplier[int] = lambda: 2
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(to_awaitable(value))

                # WHEN
                result: int = await optional.or_else_get(supplier)

                # THEN
                assert result == value

            async def test_none_value__should__return_supplied_value(self) -> None:
                """When no value is provided, return the supplied default value"""
                # GIVEN
                default_value: int = 2
                supplier: Supplier[int] = lambda: default_value
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN
                result: int = await optional.or_else_get(supplier)

                # THEN
                assert result == default_value

            async def test_async_none_value__should__return_supplied_value(self) -> None:
                """When async None value is provided, return the supplied default value"""
                # GIVEN
                default_value: int = 2
                supplier: Supplier[int] = lambda: default_value
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                result: int = await optional.or_else_get(supplier)

                # THEN
                assert result == default_value

            async def test_none_value_and_no_supplier__should__return_none_value(self) -> None:
                """When no value is provided, return the supplied default value"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN
                result: Optional[int] = await optional.or_else_get()

                # THEN
                assert result is None

            async def test_async_none_value_and_no_supplier__should__return_none_value(self) -> None:
                """When async None value is provided, return the supplied default value"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                result: Optional[int] = await optional.or_else_get()

                # THEN
                assert result is None

        class TestErrorCase:
            async def test_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When no value is provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_get(None)

            async def test_async_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When async None value is provided, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_get(None)

                await optional.or_else_get()  # Remove warning

            async def test_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When value is provided but supplier is none, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_get(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When async value is provided but supplier is none, should raise a TypeError exception"""
                # GIVEN
                value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_get(None)

                await value  # Remove warning

    class TestOrElseRaise:
        """or_else_raise method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                supplier: Supplier[Exception] = lambda: ValueError("Test error message")
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN
                result: int = await optional.or_else_raise(supplier)

                # THEN
                assert result == value

            async def test_async_value_exists__should__return_value(self) -> None:
                """When async value is provided, return the value"""
                # GIVEN
                value: int = 1
                supplier: Supplier[Exception] = lambda: ValueError("Test error message")
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(to_awaitable(value))

                # WHEN
                result: int = await optional.or_else_raise(supplier)

                # THEN
                assert result == value

            async def test_none_value__should__raise_supplied_exception(self) -> None:
                """When no value is provided, raise the supplied exception"""
                # GIVEN
                given_error: Exception = ValueError("Test error message")
                supplier: Supplier[Exception] = lambda: given_error
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(ValueError) as error:
                    await optional.or_else_raise(supplier)

                assert str(error.value) == str(given_error)

            async def test_async_none_value__should__raise_supplied_exception(self) -> None:
                """When async None value is provided, raise the supplied exception"""
                # GIVEN
                given_error: Exception = ValueError("Test error message")
                supplier: Supplier[Exception] = lambda: given_error
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN / THEN
                with raises(ValueError) as error:
                    await optional.or_else_raise(supplier)

                assert str(error.value) == str(given_error)

        class TestErrorCase:
            async def test_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When no value and no supplier are provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_raise(None)

            async def test_async_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When async None value and no supplier are provided, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_raise(None)

                await optional.or_else_get()  # Remove warning

            async def test_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When value is provided but supplier is none, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_raise(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When async value is provided but supplier is none, should raise a TypeError exception"""
                # GIVEN
                value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.or_else_raise(None)

                await value  # Remove warning

    class TestOrGet:
        """or_get method tests"""

        class TestNominalCase:
            async def test_value_exists__should__return_self_value_async_joptional(self) -> None:
                """When value is provided, return the self value AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)
                supplier: Supplier[AsyncJOptional[int]] = lambda: AsyncJOptional.of(2)

                # WHEN
                result: AsyncJOptional[int] = optional.or_get(supplier)

                # THEN
                assert await result.get() == 1

            async def test_async_value_exists__should__return_self_value_async_joptional(self) -> None:
                """When async value is provided, return the self value AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))
                supplier: Supplier[AsyncJOptional[int]] = lambda: AsyncJOptional.of(2)

                # WHEN
                result: AsyncJOptional[int] = optional.or_get(supplier)

                # THEN
                assert await result.get() == 1

            async def test_none_value__should__return_supplied_async_joptional(self) -> None:
                """When value is not provided, return the supplied default AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                other: AsyncJOptional[int] = AsyncJOptional.of(2)
                supplier: Supplier[AsyncJOptional[int]] = lambda: other

                # WHEN
                result: AsyncJOptional[int] = optional.or_get(supplier)

                # THEN
                assert await result.get() == 2

            async def test_async_none_value__should__return_supplied_async_joptional(self) -> None:
                """When provided value is async None, return the supplied default AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                other: AsyncJOptional[int] = AsyncJOptional.of(2)
                supplier: Supplier[AsyncJOptional[int]] = lambda: other

                # WHEN
                result: AsyncJOptional[int] = optional.or_get(supplier)

                # THEN
                assert await result.get() == 2

        class TestErrorCase:
            async def test_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When value is provided but supplier is not, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_get(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When async value is provided but supplier is not, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_get(None)

                await optional.or_else_get()  # Remove warning

            def test_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When no value and no supplier are provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_get(None)

            async def test_async_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When async None value and no supplier are provided, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_get(None)

                await optional.or_else_get()  # Remove warning

    class TestMap:
        """map method tests"""

        class TestNominalCase:
            async def test_value_exists__should__async_joptional_of_mapped_value(self) -> None:
                """When value is provided, should apply the given mapper"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)
                mapper: Function[int, str] = lambda i: f"Str value is: '{i}'"

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_present()
                assert await result.get() == "Str value is: '1'"

            async def test_async_value_exists__should__async_joptional_of_mapped_value(self) -> None:
                """When async value is provided, should apply the given mapper"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))
                mapper: Function[int, str] = lambda i: f"Str value is: '{i}'"

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_present()
                assert await result.get() == "Str value is: '1'"

            async def test_value_exists_with_async_mapper__should__async_joptional_of_mapped_value(self) -> None:
                """When value is provided with async mapper, should apply the given mapper"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                async def async_mapper(i: int) -> str:
                    return f"Str value is: '{i}'"

                # WHEN
                result: AsyncJOptional[str] = optional.map(async_mapper)

                # THEN
                assert await result.is_present()
                assert await result.get() == "Str value is: '1'"

            async def test_async_value_exists_with_async_mapper__should__async_joptional_of_mapped_value(self) -> None:
                """When async value is provided with async mapper, should apply the given mapper"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))

                async def async_mapper(i: int) -> str:
                    return f"Str value is: '{i}'"

                # WHEN
                result: AsyncJOptional[str] = optional.map(async_mapper)

                # THEN
                assert await result.is_present()
                assert await result.get() == "Str value is: '1'"

            async def test_value_exists_and_none_mapper_result__should__empty(self) -> None:
                """When value is provided but None mapper result, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)
                mapper: Function[int, None] = lambda i: None

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_empty()

            async def test_async_value_exists_and_none_mapper_result__should__empty(self) -> None:
                """When async value is provided but None mapper result, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))
                mapper: Function[int, None] = lambda i: None

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_empty()

            async def test_value_exists_and_async_none_mapper_result__should__empty(self) -> None:
                """When value is provided but async None mapper result, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)
                mapper: Function[int, Awaitable[None]] = lambda i: async_none()

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_empty()

            async def test_async_value_exists_and_async_none_mapper_result__should__empty(self) -> None:
                """When async value is provided but async None mapper result, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))
                mapper: Function[int, Awaitable[None]] = lambda i: async_none()

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_empty()

            async def test_none_value__should__return_empty(self) -> None:
                """When value is not provided, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                mapper: Function[int, str] = lambda i: f"Str value is: '{i}'"

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_empty()

            async def test_async_none_value__should__return_empty(self) -> None:
                """When async None value is provided, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                mapper: Function[int, str] = lambda i: f"Str value is: '{i}'"

                # WHEN
                result: AsyncJOptional[str] = optional.map(mapper)

                # THEN
                assert await result.is_empty()

        class TestErrorCase:
            async def test_value_exists_and_none_mapper__should__raise_type_error(self) -> None:
                """When value is provided but mapper is not, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.map(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_mapper__should__raise_type_error(self) -> None:
                """When async value is provided but mapper is not, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.map(None)

                await optional.or_else_get()  # Remove warning

            async def test_none_value_and_none_mapper__should__raise_type_error(self) -> None:
                """When value and mapper are not provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.map(None)

            async def test_async_none_value_and_none_mapper__should__raise_type_error(self) -> None:
                """When async none value is provided and mapper is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.map(None)

                await optional.or_else_get()  # Remove warning

    class TestFlatMap:
        """flat_map method tests"""

        class TestNominalCase:
            async def test_value_exists__should__async_joptional_of_mapped_value(self) -> None:
                """When value is provided, should apply the given mapper"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)
                mapper: Function[int, AsyncJOptional[str]] = \
                    lambda v: AsyncJOptional.of_noneable(v).map(lambda i: f"Str value is: '{i}'")

                # WHEN
                result: AsyncJOptional[str] = optional.flat_map(mapper)

                # THEN
                assert await result.is_present()
                assert await result.get() == "Str value is: '1'"

            async def test_async_value_exists__should__async_joptional_of_mapped_value(self) -> None:
                """When async value is provided, should apply the given mapper"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))
                mapper: Function[int, AsyncJOptional[str]] = \
                    lambda v: AsyncJOptional.of_noneable(v).map(lambda i: f"Str value is: '{i}'")

                # WHEN
                result: AsyncJOptional[str] = optional.flat_map(mapper)

                # THEN
                assert await result.is_present()
                assert await result.get() == "Str value is: '1'"

            async def test_none_value__should__return_empty(self) -> None:
                """When value is not provided, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                mapper: Function[int, AsyncJOptional[str]] = \
                    lambda v: AsyncJOptional.of_noneable(v).map(lambda i: f"Str value is: '{i}'")

                # WHEN
                result: AsyncJOptional[str] = optional.flat_map(mapper)

                # THEN
                assert await result.is_empty()

            async def test_async_none_value__should__return_empty(self) -> None:
                """When async None value is provided, should return an empty AsyncJOptional"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                mapper: Function[int, AsyncJOptional[str]] = \
                    lambda v: AsyncJOptional.of_noneable(v).map(lambda i: f"Str value is: '{i}'")

                # WHEN
                result: AsyncJOptional[str] = optional.flat_map(mapper)

                # THEN
                assert await result.is_empty()

        class TestErrorCase:
            async def test_value_exists_and_none_mapper__should__raise_type_error(self) -> None:
                """When value is provided but mapper is not, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_mapper__should__raise_type_error(self) -> None:
                """When async value is provided but mapper is not, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(None)

                await optional.or_else_get()  # Remove warning

            async def test_none_value_and_none_mapper__should__raise_type_error(self) -> None:
                """When value and mapper are not provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(None)

            async def test_async_none_value_and_none_mapper__should__raise_type_error(self) -> None:
                """When async None value is provided and mapper is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(None)

                await optional.or_else_get()  # Remove warning

            async def test_value_exists_and_none_mapper_result__should__raise_type_error_on_terminal_operation(
                    self
            ) -> None:
                """
                When value exists and mapper returns None, should raise a TypeError exception on terminal operation
                """
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.flat_map(lambda v: None).or_else_get()

            async def test_async_value_exists_and_none_mapper_result__should__raise_type_error_on_terminal_operation(
                    self
            ) -> None:
                """
                When async value exists and mapper returns None, should raise a TypeError exception on terminal
                operation
                """
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.flat_map(lambda v: None).or_else_get()

    class TestIfPresent:
        """if_present method tests"""

        class TestNominalCase:

            @patch('builtins.print')
            async def test_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)
                consumer: Consumer[[int]] = print

                # WHEN
                await optional.if_present(consumer)

                # THEN
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            async def test_async_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When async value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(value))
                consumer: Consumer[[int]] = print

                # WHEN
                await optional.if_present(consumer)

                # THEN
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            async def test_none_value__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When no value is provided, should do nothing"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                consumer: Consumer[[int]] = print

                # WHEN
                await optional.if_present(consumer)

                # THEN
                print_mock.assert_not_called()

            @patch('builtins.print')
            async def test_async_none_value__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When async None value is provided, should do nothing"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                consumer: Consumer[[int]] = print

                # WHEN
                await optional.if_present(consumer)

                # THEN
                print_mock.assert_not_called()

        class TestErrorCase:
            async def test_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is provided but consumer is not, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When async value is provided but consumer is not, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present(None)

                await optional.or_else_get()  # Remove warning

            async def test_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When no value and no consumer are provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present(None)

            async def test_async_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When async None value and no consumer are provided, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present(None)

                await optional.or_else_get()  # Remove warning

    class TestIfEmpty:
        """if_empty method tests"""

        class TestNominalCase:

            @patch('builtins.print')
            async def test_value_exists__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When value is provided, should do nothing"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                await optional.if_empty(runnable)

                # THEN
                print_mock.assert_not_called()

            @patch('builtins.print')
            async def test_async_value_exists__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When async value is provided, should do nothing"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(1))
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                await optional.if_empty(runnable)

                # THEN
                print_mock.assert_not_called()

            @patch('builtins.print')
            async def test_none_value__should__call_given_runnable(self, print_mock: MagicMock) -> None:
                """When no value is provided, should call the given runnable"""
                # GIVEN
                message: str = "A test message"
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                runnable: Runnable = lambda: print(message)

                # WHEN
                await optional.if_empty(runnable)

                # THEN
                print_mock.assert_called_once_with(message)

            @patch('builtins.print')
            async def test_async_none_value__should__call_given_runnable(self, print_mock: MagicMock) -> None:
                """When async None value is provided, should call the given runnable"""
                # GIVEN
                message: str = "A test message"
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                runnable: Runnable = lambda: print(message)

                # WHEN
                await optional.if_empty(runnable)

                # THEN
                print_mock.assert_called_once_with(message)

        class TestErrorCase:
            async def test_none_value_and_none_runnable__should__raise_type_error(self) -> None:
                """When no value and no runnable are provided, should raise TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_empty(cast(Runnable, None))

            async def test_async_none_value_and_none_runnable__should__raise_type_error(self) -> None:
                """When async None value and no runnable are provided, should raise TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_empty(cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_value_exists_and_none_runnable__should__raise_type_error(self) -> None:
                """When value is provided but runnable is not, should raise TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_empty(cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_runnable__should__raise_type_error(self) -> None:
                """When async value is provided but runnable is not, should do nothing"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_empty(cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

    class TestIfPresentOrElse:
        """if_present_or_else method tests"""

        class TestNominalCase:
            @patch('builtins.print')
            async def test_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)
                consumer: Consumer[[int]] = print
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                await optional.if_present_or_else(consumer, runnable)

                # THEN
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            async def test_async_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When async value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(value))
                consumer: Consumer[[int]] = print
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                await optional.if_present_or_else(consumer, runnable)

                # THEN
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            async def test_none_value__should__call_given_runnable(self, print_mock: MagicMock) -> None:
                """When value is not provided, should call the given runnable"""
                # GIVEN
                message: str = "A test message"
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                consumer: Consumer[[int]] = print
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                await optional.if_present_or_else(consumer, runnable)

                # THEN
                print_mock.assert_called_once_with(message)

            @patch('builtins.print')
            async def test_async_none_value__should__call_given_runnable(self, print_mock: MagicMock) -> None:
                """When async None value provided, should call the given runnable"""
                # GIVEN
                message: str = "A test message"
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                consumer: Consumer[[int]] = print
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                await optional.if_present_or_else(consumer, runnable)

                # THEN
                print_mock.assert_called_once_with(message)

        class TestErrorCase:
            async def test_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is provided but consumer is None, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)
                runnable: Runnable = lambda: print("A test message")

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, runnable)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When async value is provided but consumer is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)
                runnable: Runnable = lambda: print("A test message")

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, runnable)

                await optional.or_else_get()  # Remove warning

            async def test_value_exists_and_none_consumer_and_runnable__should__raise_type_error(self) -> None:
                """When value is provided but consumer and runnable are None, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_consumer_and_runnable__should__raise_type_error(self) -> None:
                """
                When async value is provided but consumer and runnable are None, should raise a TypeError exception
                """
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(async_value))

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_none_value_and_consumer_and_runnable__should__raise_type_error(self) -> None:
                """When value is not provided but consumer and runnable are None, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, cast(Runnable, None))

            async def test_async_none_value_and_consumer_and_runnable__should__raise_type_error(self) -> None:
                """
                When async None value is provided but consumer and runnable are None, should raise a TypeError exception
                """
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_none_value_runnable__should__raise_type_error(self) -> None:
                """When value is not provided and runnable is None, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                consumer: Consumer[[int]] = print

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(consumer, cast(Runnable, None))

            async def test_async_none_value_runnable__should__raise_type_error(self) -> None:
                """When async value is not provided and runnable is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)
                consumer: Consumer[[int]] = print

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(consumer, cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_value_exists_and_none_runnable__should__raise_type_error(self) -> None:
                """When value is provided but not the runnable, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)
                consumer: Consumer[[int]] = print

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(consumer, cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_runnable__should__raise_type_error(self) -> None:
                """When async value is provided but not the runnable, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(async_value))
                consumer: Consumer[[int]] = print

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(consumer, cast(Runnable, None))

                await optional.or_else_get()  # Remove warning

            async def test_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is not provided and consumer is None, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                runnable: Runnable = lambda: print("A test message")

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, runnable)

            async def test_async_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When async None value is provided and consumer is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)
                runnable: Runnable = lambda: print("A test message")

                # WHEN / THEN
                with raises(TypeError):
                    await optional.if_present_or_else(None, runnable)

                await optional.or_else_get()  # Remove warning

    class TestFilter:
        """filter method tests"""

        class TestNominalCase:
            async def test_value_with_valid_filter__should__return_async_joptional_with_value(self):
                """When value is present and predicate is valid, should return the AsyncJOptional with value"""
                # GIVEN
                value: int = 7
                predicate: Predicate[int] = lambda i: i > 5
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)

                # WHEN
                result: AsyncJOptional[int] = optional.filter(predicate)

                # THEN
                assert await result.is_present()
                assert await result.get() == value

            async def test_async_value_with_valid_filter__should__return_async_joptional_with_value(self):
                """When async value is present and predicate is valid, should return the AsyncJOptional with value"""
                # GIVEN
                value: int = 7
                predicate: Predicate[int] = lambda i: i > 5
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(value))

                # WHEN
                result: AsyncJOptional[int] = optional.filter(predicate)

                # THEN
                assert await result.is_present()
                assert await result.get() == value

            async def test_value_without_valid_filter__should__return_empty(self):
                """When value is present and predicate is not valid, should return empty AsyncJOptional"""
                # GIVEN
                value: int = 7
                predicate: Predicate[int] = lambda i: i < 5
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)

                # WHEN
                result: AsyncJOptional[int] = optional.filter(predicate)

                # THEN
                assert await result.is_empty()

            async def test_async_value_without_valid_filter__should__return_empty(self):
                """When async value is present and predicate is not valid, should return empty AsyncJOptional"""
                # GIVEN
                value: int = 7
                predicate: Predicate[int] = lambda i: i < 5
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(value))

                # WHEN
                result: AsyncJOptional[int] = optional.filter(predicate)

                # THEN
                assert await result.is_empty()

            async def test_empty_value__should__return_empty(self):
                """When value is empty, should return empty AsyncJOptional"""
                # GIVEN
                predicate: Predicate[int] = lambda i: i < 5
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN
                result: AsyncJOptional[int] = optional.filter(predicate)

                # THEN
                assert await result.is_empty()

            async def test_async_none_value__should__return_empty(self):
                """When value is async None, should return empty AsyncJOptional"""
                # GIVEN
                predicate: Predicate[int] = lambda i: i < 5
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                result: AsyncJOptional[int] = optional.filter(predicate)

                # THEN
                assert await result.is_empty()

        class TestErrorCase:
            async def test_value_exists_and_predicate_is_none__should__raise_type_error(self) -> None:
                """When value is provided but predicate is None, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.filter(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_predicate_is_none__should__raise_type_error(self) -> None:
                """When async value is provided but predicate is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.filter(None)

                await optional.or_else_get()  # Remove warning

            async def test_empty_value_and_none_predicate__should__raise_type_error(self):
                """When value is empty and predicate is None, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.filter(None)

            async def test_async_none_value_and_none_predicate__should__raise_type_error(self):
                """When value is async None and predicate is None, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.filter(None)

                await optional.or_else_get()  # Remove warning

    class TestPeek:
        """peek method tests"""

        class TestNominalCase:

            @patch('builtins.print')
            async def test_value_exists__should__call_given_consumer_on_terminal_operation(
                    self, print_mock: MagicMock
            ) -> None:
                """When value is provided, should call the given consumer on terminal operation"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(value)
                consumer: Consumer[[int]] = print

                # WHEN
                result: AsyncJOptional[int] = optional.peek(consumer)

                # THEN
                assert await result.get() == await optional.get()
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            async def test_async_value_exists__should__call_given_consumer_on_terminal_operation(
                    self, print_mock: MagicMock
            ) -> None:
                """When async value is provided, should call the given consumer on terminal operation"""
                # GIVEN
                value: int = 1
                optional: AsyncJOptional[int] = AsyncJOptional.of(to_awaitable(value))
                consumer: Consumer[[int]] = print

                # WHEN
                result: AsyncJOptional[int] = optional.peek(consumer)

                # THEN
                assert await result.get() == await optional.get()
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            async def test_none_value__should__do_nothing_on_terminal_operation(self, print_mock: MagicMock) -> None:
                """When no value is provided, should do nothing on terminal operation"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()
                consumer: Consumer[[int]] = print

                # WHEN
                result: AsyncJOptional[int] = optional.peek(consumer)

                # THEN
                assert await result.or_else_get() == await optional.or_else_get()
                print_mock.assert_not_called()

            @patch('builtins.print')
            async def test_async_none_value__should__do_nothing_on_terminal_operation(
                    self, print_mock: MagicMock
            ) -> None:
                """When async None value is provided, should do nothing on terminal operation"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_none())
                consumer: Consumer[[int]] = print

                # WHEN
                result: AsyncJOptional[int] = optional.peek(consumer)

                # THEN
                assert await result.or_else_get() == await optional.or_else_get()
                print_mock.assert_not_called()

        class TestErrorCase:
            async def test_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is provided but consumer is not, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.peek(None)

                await optional.or_else_get()  # Remove warning

            async def test_async_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When async value is provided but consumer is not, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[int] = to_awaitable(1)
                optional: AsyncJOptional[int] = AsyncJOptional.of(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.peek(None)

                await optional.or_else_get()  # Remove warning

            def test_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When no value and no consumer are provided, should raise a TypeError exception"""
                # GIVEN
                optional: AsyncJOptional[int] = AsyncJOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.peek(None)

            async def test_async_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When async None value and no consumer are provided, should raise a TypeError exception"""
                # GIVEN
                async_value: Awaitable[None] = async_none()
                optional: AsyncJOptional[int] = AsyncJOptional.of_noneable(async_value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.peek(None)

                await optional.or_else_get()  # Remove warning

    class TestToJOptional:
        """to_joptional method tests"""

        class TestNominalCase:
            async def test_empty__should__return_awaitable_none_joptional(self) -> None:
                """When empty AsyncJOptional should return empty JOptional[Awaitable[None]]"""
                # GIVEN
                async_optional: AsyncJOptional[str] = AsyncJOptional.empty()

                # WHEN
                optional: JOptional[Awaitable[str]] = async_optional.to_joptional()

                # THEN
                assert optional.is_awaitable()
                assert await optional.get() is None

            async def test_async_none_value__should__return_awaitable_none_joptional(self) -> None:
                """When async None value should return empty JOptional[Awaitable[None]]"""
                # GIVEN
                async_optional: AsyncJOptional[str] = AsyncJOptional.of_noneable(async_none())

                # WHEN
                optional: JOptional[Awaitable[str]] = async_optional.to_joptional()

                # THEN
                assert optional.is_awaitable()
                assert await optional.get() is None

            async def test_str_value__should__return_awaitable_str_joptional(self) -> None:
                """When str value should return an JOptional[Awaitable[str]] with value"""
                # GIVEN
                value: str = "A test value"
                async_optional: AsyncJOptional[str] = AsyncJOptional.of(value)

                # WHEN
                optional: JOptional[Awaitable[str]] = async_optional.to_joptional()

                # THEN
                assert optional.is_awaitable()
                assert await optional.get() == value

            async def test_async_str_value__should__return_awaitable_str_joptional(self) -> None:
                """When async str value should return an JOptional[Awaitable[str]] with value"""
                # GIVEN
                value: str = "A test value"
                async_optional: AsyncJOptional[str] = AsyncJOptional.of(to_awaitable(value))

                # WHEN
                optional: JOptional[Awaitable[str]] = async_optional.to_joptional()

                # THEN
                assert optional.is_awaitable()
                assert await optional.get() == value



"""iterable module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Iterator

from pytest import raises

from dev4py.utils import iterables


class TestGetChunks:
    """get_chunks function test"""

    class TestNominalCase:

        def test_exiting_parameters__should__return_a_chunk_generator(self) -> None:
            """When iterable and size are given, should return a chunk generator"""
            # GIVEN
            iterable: Iterable[int] = [1, 2, 3, 4, 5, 6, 7]
            chunk_size: int = 3

            # WHEN
            chunk_gen: Iterator[list[int]] = iterables.get_chunks(iterable, chunk_size)

            # THEN
            assert chunk_gen.__next__() == [1, 2, 3]
            assert chunk_gen.__next__() == [4, 5, 6]
            assert chunk_gen.__next__() == [7]
            with raises(StopIteration):
                chunk_gen.__next__()

    class TestErrorCase:
        def test_none_iterable__should__raise_type_error(self) -> None:
            """When the iterable is None, should raise TypeError exception"""
            # GIVEN
            chunk_size: int = 3

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                iterables.get_chunks(None, chunk_size).__next__()

        def test_none_chunk_size__should__raise_type_error(self) -> None:
            """When the chunk_size is None, should raise TypeError exception"""
            # GIVEN
            iterable: Iterable[int] = []

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                iterables.get_chunks(iterable, None).__next__()

        def test_negative_chunk_size__should__raise_value_error(self) -> None:
            """When the chunk_size is negative, should raise ValueError exception"""
            # GIVEN
            iterable: Iterable[int] = []
            chunk_size: int = -1

            # WHEN / THEN
            with raises(ValueError):
                # noinspection PyTypeChecker
                iterables.get_chunks(iterable, chunk_size).__next__()



"""dicts module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Any

from pytest import raises

from dev4py.utils import dicts, JOptional
from dev4py.utils.objects import is_none
from dev4py.utils.types import Supplier


class TestIsDict:
    """is_dict function tests"""

    class TestNominalCase:
        def test_none__should__return_false(self) -> None:
            """When value is None should return False"""
            # GIVEN / WHEN
            result: bool = dicts.is_dict(None)

            # THEN
            assert not result

        def test_not_dict__should__return_false(self) -> None:
            """When value is not a dict should return False"""
            # GIVEN
            value: str = "A test value"

            # WHEN
            result: bool = dicts.is_dict(value)

            # THEN
            assert not result

        def test_dict__should__return_true(self) -> None:
            """When value is a dict should return True"""
            # GIVEN
            value: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: bool = dicts.is_dict(value)

            # THEN
            assert result


class TestGetJoptionalValue:
    """get_joptional_value function tests"""

    class TestNominalCase:
        def test_none_dict__should__return_empty_joptional(self) -> None:
            """When dict is None should return empty joptional"""
            # GIVEN
            key: str = 'a key'

            # WHEN
            result: JOptional[int] = dicts.get_joptional_value(None, key)

            # THEN
            assert result.is_empty()

        def test_none_value__should__return_empty_joptional(self) -> None:
            """When value is None should return empty joptional"""
            # GIVEN
            key: str = 'c'
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: JOptional[int] = dicts.get_joptional_value(dictionary, key)

            # THEN
            assert result.is_empty()

        def test_value_exits__should__return_joptional_with_value(self) -> None:
            """When value exists should return a joptional with value"""
            # GIVEN
            key: str = 'b'
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: JOptional[int] = dicts.get_joptional_value(dictionary, key)

            # THEN
            assert result.is_present()
            assert result.get() == 2

    class TestErrorCase:
        def test_not_dict__should__raise_type_error(self) -> None:
            """When value is None should return False"""
            # GIVEN
            key: str = 'a key'
            dictionary: str = "Not a dict"

            # WHEN / THEN
            with raises(TypeError) as error:
                # noinspection PyTypeChecker
                dicts.get_joptional_value(dictionary, key)

            assert str(error.value) == "Optional[dict[K, V]] parameter is required"


class TestGetValue:
    """get_value function tests"""

    class TestNominalCase:
        def test_none_dict__should__return_none(self) -> None:
            """When dict is None should return None"""
            # GIVEN
            key: str = 'a key'

            # WHEN
            result: Optional[int] = dicts.get_value(None, key)

            # THEN
            assert result is None

        def test_none_dict_with_default_supplier__should__return_supplied_value(self) -> None:
            """When dict is None and default supplier is provided should return the supplied value"""
            # GIVEN
            key: str = 'a key'
            supplier: Supplier[int] = lambda: 2

            # WHEN
            result: Optional[int] = dicts.get_value(None, key, supplier)

            # THEN
            assert result == 2

        def test_none_value__should__return_none(self) -> None:
            """When value is None should return None"""
            # GIVEN
            key: str = 'c'
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: Optional[int] = dicts.get_value(dictionary, key)

            # THEN
            assert result is None

        def test_none_value_with_default_supplier__should__return_supplied_value(self) -> None:
            """When value is None and default supplier is provided should return the supplied value"""
            # GIVEN
            key: str = 'c'
            dictionary: dict[str, int] = {'a': 1, 'b': 2}
            supplier: Supplier[int] = lambda: 3

            # WHEN
            result: Optional[int] = dicts.get_value(dictionary, key, supplier)

            # THEN
            assert result == 3

        def test_value_exits__should__return_the_value(self) -> None:
            """When value exists should return the value"""
            # GIVEN
            key: str = 'b'
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: Optional[int] = dicts.get_value(dictionary, key)

            # THEN
            assert result == 2

        def test_value_exits_with_default_supplier__should__return_the_value(self) -> None:
            """When value exists and default supplier is provided should return the value"""
            # GIVEN
            key: str = 'b'
            dictionary: dict[str, int] = {'a': 1, 'b': 2}
            supplier: Supplier[int] = lambda: 3

            # WHEN
            result: Optional[int] = dicts.get_value(dictionary, key, supplier)

            # THEN
            assert result == 2

    class TestErrorCase:
        def test_not_dict__should__raise_type_error(self) -> None:
            """When value is None should return False"""
            # GIVEN
            key: str = 'a key'
            dictionary: str = "Not a dict"

            # WHEN / THEN
            with raises(TypeError) as error:
                # noinspection PyTypeChecker
                dicts.get_value(dictionary, key)

            assert str(error.value) == "Optional[dict[K, V]] parameter is required"


class TestGetValueFromPath:
    """get_value_from_path function tests"""

    class TestNominalCase:
        def test_none_dict__should__return_none(self) -> None:
            """When dict is None should return None"""
            # GIVEN
            path: list[str] = ['key_1', 'key_2']

            # WHEN
            result: Optional[int] = dicts.get_value_from_path(None, path)

            # THEN
            assert result is None

        def test_none_dict_with_default_supplier__should__return_supplied_value(self) -> None:
            """When dict is None and default supplier is provided should return the supplied value"""
            # GIVEN
            path: list[str] = ['key_1', 'key_2']
            supplier: Supplier[int] = lambda: 2

            # WHEN
            result: Optional[int] = dicts.get_value_from_path(None, path, supplier)

            # THEN
            assert result == 2

        def test_none_path__should__return_none(self) -> None:
            """When path is None should return None"""
            # GIVEN
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            # noinspection PyTypeChecker
            result: Optional[int] = dicts.get_value_from_path(dictionary, None)

            # THEN
            assert result is None

        def test_none_path_with_default_supplier__should__return_supplied_value(self) -> None:
            """When path is None and default supplier is provided should return the supplied value"""
            # GIVEN
            dictionary: dict[str, int] = {'a': 1, 'b': 2}
            supplier: Supplier[int] = lambda: 2

            # WHEN
            # noinspection PyTypeChecker
            result: Optional[int] = dicts.get_value_from_path(dictionary, None, supplier)

            # THEN
            assert result == 2

        def test_none_value__should__return_none(self) -> None:
            """When value is None should return None"""
            # GIVEN
            path: list[str] = ['key_1', 'key_2']
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: Optional[int] = dicts.get_value_from_path(dictionary, path)

            # THEN
            assert result is None

        def test_none_value_with_default_supplier__should__return_supplied_value(self) -> None:
            """When value is None and default supplier is provided should return the supplied value"""
            # GIVEN
            path: list[str] = ['key_1', 'key_2']
            dictionary: dict[str, int] = {'a': 1, 'b': 2}
            supplier: Supplier[int] = lambda: 3

            # WHEN
            result: Optional[int] = dicts.get_value_from_path(dictionary, path, supplier)

            # THEN
            assert result == 3

        def test_deep_dict_and_none_value__should__return_none(self) -> None:
            """When value is None in a deep dict should return None"""
            # GIVEN
            path: list[Any] = ['a', 3]
            dictionary: dict[str, dict[int, str]] = {
                'a': {1: 'a1', 2: 'a2'},
                'b': {1: 'b1', 2: 'b2'}
            }

            # WHEN
            result: Optional[str] = dicts.get_value_from_path(dictionary, path)

            # THEN
            assert result is None

        def test_deep_dict_and_none_value_with_default_supplier__should__return_supplied_value(self) -> None:
            """When value is None in a deep dict and default supplier is provided should return the supplied value"""
            # GIVEN
            path: list[Any] = ['a', 3]
            dictionary: dict[str, dict[int, str]] = {
                'a': {1: 'a1', 2: 'a2'},
                'b': {1: 'b1', 2: 'b2'}
            }
            supplier: Supplier[str] = lambda: 'a3'

            # WHEN
            result: Optional[str] = dicts.get_value_from_path(dictionary, path, supplier)

            # THEN
            assert result == 'a3'

        def test_value_exits__should__return_the_value(self) -> None:
            """When value exists should return the value"""
            # GIVEN
            path: list[str] = ['b']
            dictionary: dict[str, int] = {'a': 1, 'b': 2}

            # WHEN
            result: Optional[int] = dicts.get_value_from_path(dictionary, path)

            # THEN
            assert result == 2

        def test_value_exits_with_default_supplier__should__return_the_value(self) -> None:
            """When value exists and default supplier is provided should return the value"""
            # GIVEN
            path: list[str] = ['b']
            dictionary: dict[str, int] = {'a': 1, 'b': 2}
            supplier: Supplier[int] = lambda: 3

            # WHEN
            result: Optional[int] = dicts.get_value_from_path(dictionary, path, supplier)

            # THEN
            assert result == 2

        def test_deep_dict_and_value_exits__should__return_the_value(self) -> None:
            """When value exists in deep dict should return the value"""
            # GIVEN
            path: list[Any] = ['a', 2]
            dictionary: dict[str, dict[int, str]] = {
                'a': {1: 'a1', 2: 'a2'},
                'b': {1: 'b1', 2: 'b2'}
            }
            # WHEN
            result: Optional[str] = dicts.get_value_from_path(dictionary, path)

            # THEN
            assert result == 'a2'

        def test_deep_dict_and_value_exits_with_default_supplier__should__return_the_value(self) -> None:
            """When value exists in a deep dict and default supplier is provided should return the value"""
            # GIVEN
            path: list[Any] = ['a', 2]
            dictionary: dict[str, dict[int, str]] = {
                'a': {1: 'a1', 2: 'a2'},
                'b': {1: 'b1', 2: 'b2'}
            }
            supplier: Supplier[str] = lambda: 'a3'

            # WHEN
            result: Optional[str] = dicts.get_value_from_path(dictionary, path, supplier)

            # THEN
            assert result == 'a2'

    class TestErrorCase:
        def test_not_dict__should__raise_type_error(self) -> None:
            """When dictionary is not a dict should raise TypeError exception"""
            # GIVEN
            key: str = 'a key'
            dictionary: str = "Not a dict"

            # WHEN / THEN
            with raises(TypeError) as error:
                # noinspection PyTypeChecker
                dicts.get_value_from_path(dictionary, key)

            assert str(error.value) == "Optional[dict[K, V]] dictionary parameter must be a dict or None value"


class TestPutValue:
    """put_value function tests"""

    class TestNominalCase:
        def test_new_key__should__add_value_and_return_none(self):
            """When key doesn't exist, should add value to the key and return None"""
            # GIVEN
            key: str = 'a key'
            value: int = 1
            dictionary: dict[str, int] = {'key_1': 10}

            # WHEN
            result: Optional[int] = dicts.put_value(dictionary, key, value)

            # THEN
            assert len(dictionary) == 2
            assert dictionary[key] == value
            assert is_none(result)

        def test_key_exists__should__add_value_and_return_old_value(self):
            """When key exists, should add value to the key and return the old value"""
            # GIVEN
            key: str = 'a key'
            value: int = 1
            old_value: int = 666
            dictionary: dict[str, int] = {'key_1': 10, key: old_value}

            # WHEN
            result: Optional[int] = dicts.put_value(dictionary, key, value)

            # THEN
            assert len(dictionary) == 2
            assert dictionary[key] == value
            assert old_value == result

    class TestErrorCase:
        def test_none_dict__should__raise_type_error(self) -> None:
            """When dictionary is none should raise TypeError exception"""
            # GIVEN
            key: str = 'a key'
            value: int = 1
            dictionary: None = None

            # WHEN / THEN
            with raises(TypeError) as error:
                # noinspection PyTypeChecker
                dicts.put_value(dictionary, key, value)

            assert str(error.value) == "dictionary must be a dict value"

        def test_not_dict__should__raise_type_error(self) -> None:
            """When dictionary is not a dict should raise TypeError exception"""
            # GIVEN
            key: str = 'a key'
            value: int = 1
            dictionary: str = "Not a dict"

            # WHEN / THEN
            with raises(TypeError) as error:
                # noinspection PyTypeChecker
                dicts.put_value(dictionary, key, value)

            assert str(error.value) == "dictionary must be a dict value"


class TestEmptyDict:
    """empty_dict function tests"""

    class TestNominalCase:
        def test_should__return_an_empty_dict(self) -> None:
            """Should return an empty dict"""
            # GIVEN / WHEN
            dictionary: dict[str, int] = dicts.empty_dict()

            # THEN
            assert dictionary == {}


class TestUpdate:
    """update function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__return_first_dict_with_added_elements(self) -> None:
            """When all parameters are set, should update and return the first dict with the second one"""
            # GIVEN
            dict_1: dict[str, int] = {'k1_1': 11, 'k1_2': 12}
            dict_2: dict[str, int] = {'k2_1': 21, 'k2_2': 22}

            # WHEN
            result: dict[str, int] = dicts.update(dict_1, dict_2)

            # THEN
            expected_result: dict[str, int] = {'k1_1': 11, 'k1_2': 12, 'k2_1': 21, 'k2_2': 22}
            assert result == expected_result
            assert dict_1 == expected_result
            assert dict_2 == {'k2_1': 21, 'k2_2': 22}

        def test_empty_second_dict__should__return_first_dict_without_update(self) -> None:
            """When the second dict is empty, should return the first dict without update"""
            # GIVEN
            dict_1: dict[str, int] = {'k1_1': 11, 'k1_2': 12}
            dict_2: dict[str, int] = dicts.empty_dict()

            # WHEN
            result: dict[str, int] = dicts.update(dict_1, dict_2)

            # THEN
            expected_result: dict[str, int] = {'k1_1': 11, 'k1_2': 12}
            assert result == expected_result
            assert dict_1 == expected_result
            assert dict_2 == {}

    class TestErrorCase:
        def test_none_first_dict__should__raise_type_error(self) -> None:
            """When the first dict is None should raise TypeError exception"""
            # GIVEN
            dictionary: dict[str, int] = dicts.empty_dict()

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                dicts.update(None, dictionary)

        def test_none_second_dict__should__raise_type_error(self) -> None:
            """When the second dict is None should raise TypeError exception"""
            # GIVEN
            dictionary: dict[str, int] = dicts.empty_dict()

            # WHEN / THEN
            with raises(TypeError):
                # noinspection PyTypeChecker
                dicts.update(dictionary, None)



"""SimplePipeline module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Final, Optional

from pytest import raises

from dev4py.utils.pipeline.SimplePipeline import SimplePipeline
from dev4py.utils.types import Function


class TestSimplePipeline:
    """SimplePipeline class tests"""
    CONSTRUCTOR_ERROR_MSG: Final[str] = "SimplePipeline private constructor! Please use SimplePipeline.of"
    INIT_HANDLER: Function[int, str] = lambda v: str(v)

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:

            def test_none_create_key__should__raise_assertion_error(self):
                """When create key is none, should raise AssertionError with private constructor message"""
                # GIVEN / WHEN / THEN
                with raises(AssertionError) as error:
                    SimplePipeline(TestSimplePipeline.INIT_HANDLER, None)

                assert str(error.value) == TestSimplePipeline.CONSTRUCTOR_ERROR_MSG

            def test_invalid_create_key__should__raise_assertion_error(self):
                """When create key is invalid, should raise AssertionError with private constructor message"""
                # GIVEN / WHEN / THEN
                with raises(AssertionError) as error:
                    SimplePipeline(TestSimplePipeline.INIT_HANDLER, object())

                assert str(error.value) == TestSimplePipeline.CONSTRUCTOR_ERROR_MSG

            def test_valid_create_key__should__create_a_new_simple_pipeline(self):
                """When create key is valid, should create a new pipeline"""
                # GIVEN
                value: int = 1

                # WHEN
                # noinspection PyUnresolvedReferences
                pipeline: SimplePipeline[int, str] = SimplePipeline(TestSimplePipeline.INIT_HANDLER,
                                                                    SimplePipeline._SimplePipeline__CREATE_KEY)

                # THEN
                assert '1' == pipeline.execute(value)

            class TestErrorCase:
                def test_none_handler__should__raise_type_error(self) -> None:
                    """When no handler is provided, should raise a TypeError exception"""
                    # GIVEN / WHEN / THEN
                    with raises(TypeError):
                        # noinspection PyUnresolvedReferences
                        SimplePipeline(None, SimplePipeline._SimplePipeline__CREATE_KEY)

    class TestOf:
        """of class method tests"""

        class TestNominalCase:
            def test_handler_exists__should__return_pipeline_with_handler(self) -> None:
                """When a handler is provided, should return a SimplePipeline initialized with given handler"""
                # GIVEN
                value: int = 1

                # WHEN
                pipeline: SimplePipeline[int, str] = SimplePipeline.of(TestSimplePipeline.INIT_HANDLER)

                # THEN
                assert '1' == pipeline.execute(value)

        class TestErrorCase:
            def test_none_handler__should__raise_type_error(self) -> None:
                """When no handler is provided, should raise a TypeError exception"""
                # GIVEN / WHEN / THEN
                with raises(TypeError):
                    SimplePipeline.of(None)

    class TestAddHandler:
        """add_handler method tests"""

        class TestNominalCase:
            def test_handler_exists__should__return_pipeline_with_added_handler(self) -> None:
                """
                When a handler is provided, should return a new SimplePipeline with given handler(/operation) added
                """
                # GIVEN
                value: int = 1
                pipeline: SimplePipeline[int, str] = SimplePipeline.of(TestSimplePipeline.INIT_HANDLER)
                handler: Function[str, str] = lambda s: s + "_new_handler_suffix"

                # WHEN
                new_pipeline: SimplePipeline[int, str] = pipeline.add_handler(handler)

                # THEN
                assert '1_new_handler_suffix' == new_pipeline.execute(value)

        class TestErrorCase:
            def test_none_handler__should__raise_type_error(self) -> None:
                """When no handler is provided, should raise a TypeError exception"""
                # GIVEN
                pipeline: SimplePipeline[int, str] = SimplePipeline.of(TestSimplePipeline.INIT_HANDLER)

                # WHEN / THEN
                with raises(TypeError):
                    pipeline.add_handler(None)

    class TestExecute:
        """execute method tests"""

        class TestNominalCase:
            def test_none_value__should__return_execution_result(self) -> None:
                """When a value is None, should return the execution result"""
                # GIVEN
                pipeline: SimplePipeline[Optional[int], str] = SimplePipeline.of(TestSimplePipeline.INIT_HANDLER)

                # WHEN
                result: str = pipeline.execute(None)

                # THEN
                assert 'None' == result

            def test_value_exists__should__return_execution_result(self) -> None:
                """When a value is provided, should return the execution result"""
                # GIVEN
                value: int = 1
                pipeline: SimplePipeline[int, str] = SimplePipeline.of(TestSimplePipeline.INIT_HANDLER)

                # WHEN
                result: str = pipeline.execute(value)

                # THEN
                assert '1' == result



"""collectors module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import FrozenInstanceError, dataclass
from typing import Final

from pytest import raises

from dev4py.utils import Collector, collectors
from dev4py.utils.objects import non_none, to_self
from dev4py.utils.types import Supplier, BiFunction, BiConsumer, Function


class TestCollector:
    """Collector class tests"""

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:

            def test_existing_parameters__should__create_a_new_collector(self) -> None:
                """When all parameters are set, should create a new collector with given parameters"""
                # GIVEN
                supplier: Final[Supplier[int]] = lambda: 0
                accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

                # WHEN
                collector: Collector = Collector(supplier=supplier, accumulator=accumulator, combiner=combiner)

                # THEN
                assert non_none(collector)
                assert collector.supplier == supplier
                assert collector.accumulator == accumulator
                assert collector.combiner == combiner

        class TestErrorCase:
            def test_none_supplier__should__raise_type_error(self) -> None:
                """When supplier is None should raise a TypeError exception"""
                # GIVEN
                accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

                # WHEN / THEN
                with raises(TypeError):
                    Collector(supplier=None, accumulator=accumulator, combiner=combiner)

            def test_none_accumulator__should__raise_type_error(self) -> None:
                """When accumulator is None should raise a TypeError exception"""
                # GIVEN
                supplier: Final[Supplier[int]] = lambda: 0
                combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

                # WHEN / THEN
                with raises(TypeError):
                    Collector(supplier=supplier, accumulator=None, combiner=combiner)

            def test_none_combiner__should__raise_type_error(self) -> None:
                """When combiner is None should raise a TypeError exception"""
                # GIVEN
                supplier: Final[Supplier[int]] = lambda: 0
                accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

                # WHEN / THEN
                with raises(TypeError):
                    Collector(supplier=supplier, accumulator=accumulator, combiner=None)

    class TestSetters:
        """Setters tests"""

        class TestErrorCase:
            def test_set_new_supplier__should__raise_frozen_instance_error(self) -> None:
                """When set a new supplier, should raise a FrozenInstanceError exception"""
                # GIVEN
                supplier: Final[Supplier[int]] = lambda: 0
                accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                collector: Collector = Collector(supplier=supplier, accumulator=accumulator, combiner=combiner)

                # WHEN / THEN
                new_supplier: Final[Supplier[int]] = lambda: 10
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    collector.supplier = new_supplier

            def test_set_new_accumulator__should__raise_frozen_instance_error(self) -> None:
                """When set a new accumulator, should raise a FrozenInstanceError exception"""
                # GIVEN
                supplier: Final[Supplier[int]] = lambda: 0
                accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                collector: Collector = Collector(supplier=supplier, accumulator=accumulator, combiner=combiner)

                # WHEN / THEN
                new_accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2 + 2
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    collector.accumulator = new_accumulator

            def test_set_new_combiner__should__raise_frozen_instance_error(self) -> None:
                """When set a new combiner, should raise a FrozenInstanceError exception"""
                # GIVEN
                supplier: Final[Supplier[int]] = lambda: 0
                accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
                collector: Collector = Collector(supplier=supplier, accumulator=accumulator, combiner=combiner)

                # WHEN / THEN
                new_combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2 + 2
                with raises(FrozenInstanceError):
                    # noinspection PyDataclass
                    collector.combiner = new_combiner


class TestOf:
    """of function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__create_a_new_collector(self) -> None:
            """When all parameters are set, should create a new collector with given parameters"""
            # GIVEN
            supplier: Final[Supplier[int]] = lambda: 0
            accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
            combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

            # WHEN
            collector: Collector = collectors.of(supplier=supplier, accumulator=accumulator, combiner=combiner)

            # THEN
            assert non_none(collector)
            assert collector.supplier == supplier
            assert collector.accumulator == accumulator
            assert collector.combiner == combiner

    class TestErrorCase:
        def test_none_supplier__should__raise_type_error(self) -> None:
            """When supplier is None should raise a TypeError exception"""
            # GIVEN
            accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2
            combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

            # WHEN / THEN
            with raises(TypeError):
                collectors.of(supplier=None, accumulator=accumulator, combiner=combiner)

        def test_none_accumulator__should__raise_type_error(self) -> None:
            """When accumulator is None should raise a TypeError exception"""
            # GIVEN
            supplier: Final[Supplier[int]] = lambda: 0
            combiner: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

            # WHEN / THEN
            with raises(TypeError):
                collectors.of(supplier=supplier, accumulator=None, combiner=combiner)

        def test_none_combiner__should__raise_type_error(self) -> None:
            """When combiner is None should raise a TypeError exception"""
            # GIVEN
            supplier: Final[Supplier[int]] = lambda: 0
            accumulator: Final[BiFunction[int, int, int]] = lambda i1, i2: i1 + i2

            # WHEN / THEN
            with raises(TypeError):
                collectors.of(supplier=supplier, accumulator=accumulator, combiner=None)


@dataclass
class IntClass:
    value: int = 0

    def add(self, value: int) -> None:
        self.value += value


class TestOfBiconsumers:
    """of_biconsumers function tests"""

    class TestNominalCase:
        def test_existing_parameters__should__create_a_new_collector(self) -> None:
            """When all parameters are set, should create a new collector with given parameters"""
            # GIVEN
            supplier: Final[Supplier[IntClass]] = lambda: IntClass()
            accumulator: Final[BiConsumer[IntClass, int]] = lambda icls, i: icls.add(i)
            combiner: Final[BiConsumer[IntClass, IntClass]] = lambda icls1, icls2: icls1.add(icls2.value)

            # WHEN
            collector: Collector = collectors \
                .of_biconsumers(supplier=supplier, accumulator=accumulator, combiner=combiner)

            # THEN
            assert non_none(collector)
            collector_supplier: Final[Supplier[IntClass]] = collector.supplier
            assert collector_supplier == supplier
            assert collector_supplier().value == 0
            assert collector.accumulator(IntClass(value=2), 8).value == 10
            assert collector.combiner(IntClass(value=2), IntClass(value=8)).value == 10

    class TestErrorCase:
        def test_none_supplier__should__raise_type_error(self) -> None:
            """When supplier is None should raise a TypeError exception"""
            # GIVEN
            accumulator: Final[BiConsumer[IntClass, int]] = lambda icls, i: icls.add(i)
            combiner: Final[BiConsumer[IntClass, IntClass]] = lambda icls1, icls2: icls1.add(icls2.value)

            # WHEN / THEN
            with raises(TypeError):
                collectors.of_biconsumers(supplier=None, accumulator=accumulator, combiner=combiner)

        def test_none_accumulator__should__raise_type_error(self) -> None:
            """When accumulator is None should raise a TypeError exception"""
            # GIVEN
            supplier: Final[Supplier[IntClass]] = lambda: IntClass()
            combiner: Final[BiConsumer[IntClass, IntClass]] = lambda icls1, icls2: icls1.add(icls2.value)

            # WHEN / THEN
            with raises(TypeError):
                collectors.of_biconsumers(supplier=supplier, accumulator=None, combiner=combiner)

        def test_none_combiner__should__raise_type_error(self) -> None:
            """When combiner is None should raise a TypeError exception"""
            # GIVEN
            supplier: Final[Supplier[IntClass]] = lambda: IntClass()
            accumulator: Final[BiConsumer[IntClass, int]] = lambda icls, i: icls.add(i)

            # WHEN / THEN
            with raises(TypeError):
                collectors.of_biconsumers(supplier=supplier, accumulator=accumulator, combiner=None)


class TestToList:
    """to_list function tests"""

    class TestNominalCase:
        def test_should__return_list_collector(self) -> None:
            """Should return a Collector that accumulates the input elements into a new list"""
            # GIVEN / WHEN
            collector: Collector[int, list[int]] = collectors.to_list()

            # THEN
            assert collector.supplier() == []
            assert collector.accumulator([], 1) == [1]
            assert collector.combiner([1], [2]) == [1, 2]


class TestToDict:
    """to_dict function tests"""

    class TestNominalCase:
        def test_existing_mappers__should__return_dict_collector(self) -> None:
            """When key/value mappers are provided, should return a dict collector"""
            # GIVEN
            key_mapper: Function[int, str] = lambda i: str(i)
            value_mapper: Function[int, int] = to_self

            # WHEN
            collector: Collector[int, dict[str, int]] = collectors.to_dict(key_mapper, value_mapper)

            # THEN
            assert collector.supplier() == {}
            assert collector.accumulator({}, 1) == {'1': 1}
            assert collector.combiner({'1': 1}, {'2': 2}) == {'1': 1, '2': 2}

    class TestErrorCase:
        def test_none_key_mapper__should__raise_type_error(self) -> None:
            """When key mapper is None should raise a TypeError exception"""
            # GIVEN
            value_mapper: Function[int, int] = to_self

            # WHEN / THEN
            with raises(TypeError):
                collectors.to_dict(None, value_mapper)

        def test_none_value_mapper__should__raise_type_error(self) -> None:
            """When value mapper is None should raise a TypeError exception"""
            # GIVEN
            key_mapper: Function[int, str] = lambda i: str(i)

            # WHEN / THEN
            with raises(TypeError):
                collectors.to_dict(key_mapper, None)


class TestToNone:
    """to_none function tests"""

    class TestNominalCase:
        def test_should__return_none_collector(self) -> None:
            """Should return a Collector that always returns None"""
            # GIVEN / WHEN
            collector: Collector[int, None] = collectors.to_none()

            # THEN
            assert collector.supplier() is None
            assert collector.accumulator(None, None) is None
            assert collector.accumulator(None, 1) is None
            assert collector.accumulator([], None) is None
            assert collector.accumulator([], 1) is None
            assert collector.combiner(None, None) is None
            assert collector.combiner([1], None) is None
            assert collector.combiner(None, [2]) is None
            assert collector.combiner([1], [2]) is None


class TestToTuple:
    """to_tuple function tests"""

    class TestNominalCase:
        def test_should__return_tuple_collector(self) -> None:
            """Should return a Collector that accumulates the input elements into a new tuple"""
            # GIVEN / WHEN
            collector: Collector[int, tuple[int, ...]] = collectors.to_tuple()

            # THEN
            assert collector.supplier() == ()
            assert collector.accumulator((), 1) == (1,)
            assert collector.combiner((1,), (2,)) == (1, 2)


class TestToCounter:
    """to_counter function tests"""

    class TestNominalCase:
        def test_should__return_counter_collector(self) -> None:
            """Should return a Collector accepting elements of type T that counts the number of input elements"""
            # GIVEN / WHEN
            collector: Collector[int, int] = collectors.to_counter()

            # THEN
            assert collector.supplier() == 0
            assert collector.accumulator(1, 10) == 2
            assert collector.combiner(3, 7) == 10



"""
The `objects` module provides a set of utility functions to simplify objects/variables operations or checks
"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, cast, Awaitable

from dev4py.utils.awaitables import is_awaitable
from dev4py.utils.types import Supplier, T, SyncOrAsync


def is_none(obj: Any) -> bool:
    """
    Checks if the given object is None

    Args:
        obj: The object to check

    Returns:
        object: True if obj is None, False otherwise
    """
    return obj is None


def non_none(obj: Any) -> bool:
    """
    Checks if the given object is not None

    Args:
        obj: The object to check

    Returns:
        object: True if obj is NOT None, False otherwise
    """
    return obj is not None


def require_non_none(obj: Optional[T], message: str = "None object error") -> T:
    """
    Checks if the given object is not None or raises an error

    Args:
        obj: The object to check
        message: The error message is case of obj is None

    Returns:
        object: obj if obj is not None

    Raises:
        TypeError: Raises a TypeError if obj is None
    """
    if is_none(obj):
        raise TypeError(message)
    return cast(T, obj)


def require_non_none_else(obj: Optional[T], default: T) -> T:
    """
    Checks if the given object is not None and returns it or returns the default one

    Args:
        obj: The object to check and return
        default: The default object to return if obj is None

    Returns:
        object: obj if not None, default_obj otherwise

    Raises:
        TypeError: Raises a TypeError if obj and default_obj are None
    """
    return cast(T, obj) if non_none(obj) else require_non_none(default)


def require_non_none_else_get(obj: Optional[T], supplier: Supplier[T]) -> T:
    """
    Checks if the given object is not None and returns it or returns the object from the given supplier

    Args:
        obj: The object to check and return
        supplier: The supplier to call if obj is None

    Returns:
        object: obj if not None, the supplier call result otherwise

    Raises:
        TypeError: Raises a TypeError if obj and supplier or supplied object are None
    """
    return cast(T, obj) if non_none(obj) else require_non_none(require_non_none(supplier, "Supplier cannot be None")())


def to_string(obj: Any, default_str: Optional[str] = None) -> str:
    """
    Returns the result of calling str function for the given object if not null or default_str otherwise

    Args:
        obj: The object to stringify
        default_str: The str value to return if obj is None

    Returns:
        str: A str representing obj, default_str otherwise
    """
    return str(obj if non_none(obj) else default_str)


async def async_require_non_none(obj: SyncOrAsync[T], message: str = "None async object error") -> T:
    """
    Checks if the given object or awaitable object result is not None or raises an error

    Args:
        obj: The object or awaitable object to check
        message: The error message is case of obj or awaitable obj is None

    Returns:
        object: obj if obj is not None and not an Awaitable or Awaitable result if obj is an awaitable and the result is
        not None

    Raises:
        TypeError: Raises a TypeError if obj is None or if obj is an awaitable with None result
    """
    return require_non_none(await cast(Awaitable[T], obj) if is_awaitable(obj) else cast(T, obj), message)


async def async_require_non_none_else(obj: SyncOrAsync[Optional[T]], default: T) -> T:
    """
    Checks if the given object or awaitable object result is not None and returns it or returns the default one

    Args:
        obj: The object or awaitable object to check
        default: The default object to return if obj is None

    Returns:
        object:obj if obj is not None and not an Awaitable or Awaitable result if obj is an awaitable and the result is
        not None, default_obj otherwise

    Raises:
        TypeError: Raises a TypeError if obj (or await obj) and default_obj are None
    """
    return require_non_none_else(
        await cast(Awaitable[Optional[T]], obj) if is_awaitable(obj) else cast(Optional[T], obj),
        default
    )


async def async_require_non_none_else_get(obj: SyncOrAsync[Optional[T]], supplier: Supplier[T]) -> T:
    """
    Checks if the given object or awaitable object result is not None and returns it or returns the object from the
    given supplier

    Args:
        obj: The object or awaitable object to check
        supplier: The supplier to call if obj or awaitable obj result is None

    Returns:
        object: obj if not None and not an Awaitable or Awaitable result if obj is an awaitable and the result is
        not None, the supplier call result otherwise

    Raises:
        TypeError: Raises a TypeError if obj (or await obj)  and supplier or supplied object are None
    """
    return require_non_none_else_get(
        await cast(Awaitable[Optional[T]], obj) if is_awaitable(obj) else cast(Optional[T], obj),
        supplier
    )


def to_none(*args: Any, **kwargs: Any) -> None:  # pylint: disable=W0613
    """
    Returns None whatever the parameters

    Args:
        *args: positional parameters
        **kwargs: named parameters

    Returns:
        None: None whatever the parameters
    """
    return None


def to_self(obj: T) -> T:
    """
    Returns the given parameter
    Note: can be useful with multiprocessing where lambda cannot be used (lambda are not serializable)

    Args:
        obj: The object to return

    Returns:
        obj (T): The given parameter
    """
    return obj



"""The `JOptional` class provides a java like Optional class"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import partial
from typing import Generic, Optional, Final, Any, cast, Awaitable

from dev4py import utils
from dev4py.utils import objects
from dev4py.utils.awaitables import is_awaitable
from dev4py.utils.objects import to_self, to_none
from dev4py.utils.types import T, Supplier, Function, R, Consumer, Runnable, Predicate, SyncOrAsync


class JOptional(Generic[T]):
    """A class inspired by the java.util.Optional class"""

    __EMPTY: JOptional[Any]
    __CREATE_KEY: Final[object] = object()
    __NO_VALUE_ERROR_MSG: Final[str] = "No value present"

    @classmethod
    def of(cls, value: T) -> JOptional[T]:
        """
        Returns a JOptional describing the given non-None value

        Args:
            value: The value to describe, which must be non-None

        Returns:
            JOptional[T]: A JOptional of value T type with the value present

        Raises:
            TypeError: Raises a TypeError if value is None
        """
        return JOptional(objects.require_non_none(value, cls.__NO_VALUE_ERROR_MSG), cls.__CREATE_KEY)

    @classmethod
    def of_noneable(cls, value: Optional[T]) -> JOptional[T]:
        """
        Returns a JOptional describing the given value, if non-None, otherwise returns an empty JOptional

        Args:
            value: The possibly-None value to describe

        Returns:
            JOptional[T]: A JOptional of value T type with the value present if the specified value is non-None,
            otherwise an empty JOptional
        """
        return cls.empty() if objects.is_none(value) else cls.of(cast(T, value))

    @classmethod
    def empty(cls) -> JOptional[T]:
        """Return an empty instance of JOptional"""
        return cls.__EMPTY

    def __init__(self, value: Optional[T], create_key: object):
        """JOptional private constructor: Constructs an instance with the described value"""
        assert create_key == self.__CREATE_KEY, \
            "JOptional private constructor! Please use JOptional.of or JOptional.of_noneable"
        self._value: Optional[T] = value

    def is_present(self) -> bool:
        """
        If a value is present, returns true, otherwise false

        Returns:
            bool: true if a value is present, otherwise false
        """
        return objects.non_none(self._value)

    def is_empty(self) -> bool:
        """
        If a value is not present, returns true, otherwise false.

        Returns:
            bool: true if a value is not present, otherwise false
        """
        return objects.is_none(self._value)

    def get(self) -> T:
        """
        If a value is present, returns the value, otherwise raises a ValueError

        Returns:
            value: The non-None value described by this JOptional

        Raises:
            ValueError: Raises a ValueError if no value is present
        """
        return self.or_else_raise()

    def or_else(self, other: Optional[T] = None) -> Optional[T]:
        """
        If a value is present, returns the value, otherwise returns other
        Args:
            other: The value to be returned, if no value is present. May be None.

        Returns:
            value: The value, if present, otherwise other
        """
        return self.or_else_get(cast(Supplier[Optional[T]], partial(to_self, obj=other)))

    def or_else_get(self, supplier: Supplier[Optional[T]] = to_none) -> Optional[T]:
        """
        If a value is present, returns the value, otherwise returns the result produced by the supplying function

        Args:
            supplier: The supplying function that produces a value to be returned

        Returns:
            value: The value, if present, otherwise the result produced by the supplying function

        Raises:
            TypeError: if the supplying function is None
        """
        objects.require_non_none(supplier)
        return self._value if self.is_present() else supplier()

    @staticmethod  # pragma: no mutate
    def __or_else_raise_supplier_lambda() -> Exception:
        """
        private static method to replace or_else_raise lambda supplier by a function
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        # lambda: cast(Exception, ValueError(JOptional.__NO_VALUE_ERROR_MSG))
        return cast(Exception, ValueError(JOptional.__NO_VALUE_ERROR_MSG))

    def or_else_raise(
            self, supplier: Supplier[Exception] = __or_else_raise_supplier_lambda
    ) -> T:
        """
        If a value is present, returns the value, otherwise raises an exception produced by the exception supplying
        function

        Args:
            supplier: The supplying function that produces an exception to be raised (default: ValueError)

        Returns:
            value: The value, if present

        Raises:
            Exception: if the supplying function is None
            TypeError: if supplier is None
        """
        objects.require_non_none(supplier)
        if self.is_empty():
            raise supplier()
        return cast(T, self._value)

    def or_get(self, supplier: Supplier[JOptional[T]]) -> JOptional[T]:
        """
        This is the `or` equivalent in java Optional (Reminder: `or` is a python keyword)

        If a value is present, returns a JOptional describing the value, otherwise returns a JOptional produced by the
        supplying function

        Args:
            supplier: The supplying function that produces a JOptional to be returned

        Returns:
            JOptional[T]: Returns a JOptional describing the value of this JOptional, if a value is present, otherwise
            a JOptional produced by the supplying function

        Raises:
            TypeError: if the supplying function is None
        """
        objects.require_non_none(supplier)
        return self if self.is_present() else objects.require_non_none(supplier())

    @staticmethod  # pragma: no mutate
    def __map_lambda(v: T, mapper: Function[T, Optional[R]]) -> JOptional[R]:
        """
        private static method to replace inner map lambda supplier by a function
        Note: lambda are not used in order to be compatible with multiprocessing (lambda are not serializable)
        """
        # lambda v: JOptional.of_noneable(mapper(v))
        return JOptional.of_noneable(mapper(v))

    def map(self, mapper: Function[T, Optional[R]]) -> JOptional[R]:
        """
        If a value is present, returns a JOptional describing (as if by of_noneable) the result of applying the given
        mapping function to the value, otherwise returns an empty JOptional

        If the mapping function returns a None result then this method returns an empty JOptional.

        Args:
            mapper: The mapping function to apply to a value, if present

        Returns:
            JOptional[R]: A JOptional describing the result of applying a mapping function to the value of this
            JOptional, if a value is present, otherwise an empty JOptional

        Raises:
            TypeError: If the mapping function is None
        """
        objects.require_non_none(mapper)
        return self.flat_map(partial(JOptional.__map_lambda, mapper=mapper))

    def flat_map(self, mapper: Function[T, JOptional[R]]) -> JOptional[R]:
        """
        If a value is present, returns the result of applying the given JOptional-bearing mapping function to the value,
        otherwise returns an empty JOptional

        This method is similar to map(Function), but the mapping function is one whose result is already a JOptional,
        and if invoked, flatMap does not wrap it within an additional JOptional

        Args:
            mapper: The mapping function to apply to a value, if present

        Returns:
            JOptional[R]: The result of applying a JOptional-bearing mapping function to the value of this JOptional,
            if a value is present, otherwise an empty JOptional

        Raises:
            TypeError: if the mapping function is None
        """
        objects.require_non_none(mapper)
        return JOptional.empty() if self.is_empty() else objects.require_non_none(mapper(cast(T, self._value)))

    def if_present(self, consumer: Consumer[T]) -> None:
        """
        If a value is present, performs the given consumer with the value, otherwise does nothing

        Args:
            consumer: The consumer to be performed, if a value is present

        Returns:
            Nothing

        Raises:
            TypeError: if the consumer is None
        """
        objects.require_non_none(consumer)
        # pylint: disable=W0106
        self.is_present() and consumer(cast(T, self._value))

    def if_empty(self, empty_action: Runnable) -> None:
        """
        If a value is not present, performs the given runnable, otherwise does nothing

        Args:
            empty_action: The runnable to be performed, if a value is not present

        Returns:
            Nothing

        Raises:
            TypeError: if the runnable is None
        """
        objects.require_non_none(empty_action)
        # pylint: disable=W0106
        self.is_empty() and empty_action()

    def if_present_or_else(self, consumer: Consumer[T], empty_action: Runnable) -> None:
        """
        If a value is present, performs the given consumer with the value, otherwise performs the given empty_action

        Args:
            consumer: The consumer to be performed, if a value is present
            empty_action: The runnable to be performed, if a value is not

        Returns:
            Nothing

        Raises:
            TypeError: if the consumer is None or if the runnable is None
        """
        objects.require_non_none(consumer)
        objects.require_non_none(empty_action)
        # pylint: disable=W0106
        consumer(cast(T, self._value)) if self.is_present() else empty_action()

    def filter(self, predicate: Predicate[T]) -> JOptional[T]:
        """
        If a value is present, and the value matches the given predicate, returns a JOptional describing the value,
        otherwise returns an empty JOptional

        Args:
            predicate: The predicate to apply to a value, if present

        Returns:
            JOptional[T]: A JOptional describing the value of this JOptional, if a value is present and the value
            matches the given predicate, otherwise an empty JOptional

        Raises:
            TypeError: if the predicate is None
        """
        objects.require_non_none(predicate)
        return self if self.is_empty() or predicate(cast(T, self._value)) else JOptional.empty()

    def peek(self, consumer: Consumer[T]) -> JOptional[T]:
        """
        If a value is present, performs the given consumer with the value, otherwise does nothing and returns the
        current JOptional

        Args:
            consumer: The consumer to be performed, if a value is present

        Returns:
             JOptional[T]: The current JOptional (self)

        Raises:
            TypeError: if the consumer is None
        """
        objects.require_non_none(consumer)
        self.if_present(consumer)
        return self

    def is_awaitable(self) -> bool:
        """
        If the value is an Awaitable (Coroutine, Task or Future), returns true, otherwise false

        Returns:
            bool: true if the value is an Awaitable, otherwise false
        """
        return is_awaitable(self._value)

    async def to_sync_value(self) -> JOptional[Any]:
        """
        If the value is an Awaitable (Coroutine, Task or Future), awaits and returns a JOptional with the obtained
        value, otherwise if the value is not an Awaitable returns self

        Note: can be useful in order to call is_present, is_empty, if_present, if_empty, ect. depending on the awaited
        value result (otherwise, for example, is_present is always True with an 'Awaitable' value)

        Returns:
            JOptional: A JOptional with synchronized value (i.e. not an Awaitable value)
        """
        return JOptional.of_noneable(await cast(Awaitable[Any], self._value)) if self.is_awaitable() else self

    def to_async_joptional(self) -> utils.AsyncJOptional[SyncOrAsync[T]]:
        """
        Convert the current JOptional to an AsyncJOptional

        Note: It can be useful in order to use async mapper

        Returns:
            AsyncJOptional[T]: The corresponding AsyncJOptional
        """
        return utils.AsyncJOptional.of_noneable(self._value)


# INIT STATIC VARIABLES
# noinspection PyProtectedMember
# noinspection PyUnresolvedReferences
# pylint: disable=W0212
JOptional._JOptional__EMPTY = JOptional(None, JOptional._JOptional__CREATE_KEY)  # type: ignore



"""JOptional module tests"""

# Copyright 2022 the original author or authors (i.e.: St4rG00se for Dev4py).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from asyncio import Task, Future
from typing import Final, Optional, cast, Awaitable
from unittest.mock import patch, MagicMock

from pytest import raises

from dev4py.utils import JOptional, AsyncJOptional
from dev4py.utils.awaitables import to_awaitable
from dev4py.utils.types import Supplier, Function, Consumer, Runnable, Predicate


class TestJOptional:
    """JOptional class tests"""
    NO_VALUE_ERROR_MSG: Final[str] = "No value present"
    CONSTRUCTOR_ERROR_MSG: Final[str] = \
        "JOptional private constructor! Please use JOptional.of or JOptional.of_noneable"

    class TestConstructor:
        """Constructor tests"""

        class TestNominalCase:

            def test_none_create_key__should__raise_assertion_error(self):
                """When create key is none, should raise AssertionError with private constructor message"""
                # GIVEN / WHEN / THEN
                with raises(AssertionError) as error:
                    JOptional("A value", None)

                assert str(error.value) == TestJOptional.CONSTRUCTOR_ERROR_MSG

            def test_invalid_create_key__should__raise_assertion_error(self):
                """When create key is invalid, should raise AssertionError with private constructor message"""
                # GIVEN / WHEN / THEN
                with raises(AssertionError) as error:
                    JOptional("A value", object())

                assert str(error.value) == TestJOptional.CONSTRUCTOR_ERROR_MSG

            def test_valid_create_key__should__create_a_new_joptional(self):
                """When create key is valid, should create a new JOptional"""
                # GIVEN
                value: int = 1

                # WHEN
                # noinspection PyUnresolvedReferences
                optional: JOptional[int] = JOptional(value, JOptional._JOptional__CREATE_KEY)

                # THEN
                assert optional.is_present()
                assert optional._value == value

    class TestOf:
        """of class method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_joptional_with_value(self) -> None:
                """When a value is provided, should return a joptional with value"""
                # GIVEN
                value: int = 1

                # WHEN
                optional: JOptional[int] = JOptional.of(value)

                # THEN
                assert optional._value == value
                assert optional.is_present()

        class TestErrorCase:
            def test_none_value__should__raise_type_error(self) -> None:
                """When no value is provided, should raise a TypeError exception with no value message"""
                # GIVEN / WHEN / THEN
                with raises(TypeError) as error:
                    JOptional.of(None)

                assert str(error.value) == TestJOptional.NO_VALUE_ERROR_MSG

    class TestOfNoneable:
        """of_noneable class method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_joptional_with_value(self) -> None:
                """When a value is provided, should return a joptional with value"""
                # GIVEN
                value: int = 1

                # WHEN
                optional: JOptional[int] = JOptional.of_noneable(value)

                # THEN
                assert optional._value == value
                assert optional.is_present()

            def test_none_value__should__return_empty_joptional(self) -> None:
                """When no value is provided, should return an empty joptional"""
                # GIVEN
                value: Optional[int] = None

                # WHEN
                optional: JOptional[int] = JOptional.of_noneable(value)

                # THEN
                assert optional._value == value
                assert optional.is_empty()

    class TestEmpty:
        """empty class method tests"""

        class TestNominalCase:
            def test_should__return_empty_joptional(self) -> None:
                """Should return a joptional with value"""
                # GIVEN / WHEN
                optional: JOptional[int] = JOptional.empty()

                # THEN
                assert optional._value is None
                assert optional.is_empty()

    class TestIsPresent:
        """is_present method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_true(self) -> None:
                """When a value is provided, should return true"""
                # GIVEN / WHEN
                optional: JOptional[int] = JOptional.of(1)

                # THEN
                assert optional.is_present()

            def test_none_value__should__return_false(self) -> None:
                """When no value is provided, should return false"""
                # GIVEN / WHEN
                optional: JOptional[int] = JOptional.empty()

                # THEN
                assert not optional.is_present()

    class TestIsEmpty:
        """is_empty method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_false(self) -> None:
                """When a value is provided, should return false"""
                # GIVEN / WHEN
                optional: JOptional[int] = JOptional.of(1)

                # THEN
                assert not optional.is_empty()

            def test_none_value__should__return_true(self) -> None:
                """When no value is provided, should return true"""
                # GIVEN / WHEN
                optional: JOptional[int] = JOptional.empty()

                # THEN
                assert optional.is_empty()

    class TestGet:
        """get method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of_noneable(value)

                # WHEN
                result: int = optional.get()

                # THEN
                assert result == value

        class TestErrorCase:
            def test_none_value__should__raise_type_error(self) -> None:
                """When no value is provided, should raise a ValueError exception with no value message"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(ValueError) as error:
                    optional.get()

                assert str(error.value) == TestJOptional.NO_VALUE_ERROR_MSG

    class TestOrElse:
        """or_else method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                default_value: int = 2
                optional: JOptional[int] = JOptional.of_noneable(value)

                # WHEN
                result: int = optional.or_else(default_value)

                # THEN
                assert result == value

            def test_none_value__should__return_default_value(self) -> None:
                """When no value is provided, return the default value"""
                # GIVEN
                default_value: int = 2
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: int = optional.or_else(default_value)

                # THEN
                assert result == default_value

            def test_none_value_and_none_default__should__return_none(self) -> None:
                """When no value is provided and default value is None, return the default value"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: Optional[int] = optional.or_else(None)

                # THEN
                assert result is None

            def test_none_value_and_no_default__should__return_none(self) -> None:
                """When no value is provided and no default value, return None value"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: Optional[int] = optional.or_else()

                # THEN
                assert result is None

    class TestOrElseGet:
        """or_else_get method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                supplier: Supplier[int] = lambda: 2
                optional: JOptional[int] = JOptional.of_noneable(value)

                # WHEN
                result: int = optional.or_else_get(supplier)

                # THEN
                assert result == value

            def test_none_value__should__return_supplied_value(self) -> None:
                """When no value is provided, return the supplied default value"""
                # GIVEN
                default_value: int = 2
                supplier: Supplier[int] = lambda: default_value
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: int = optional.or_else_get(supplier)

                # THEN
                assert result == default_value

            def test_none_value_and_no_supplier__should__return_none_value(self) -> None:
                """When no value is provided, return the supplied default value"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: Optional[int] = optional.or_else_get()

                # THEN
                assert result is None

        class TestErrorCase:
            def test_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When no value is provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_else_get(None)

            def test_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When value is provided but supplier is none, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of_noneable(value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_else_get(None)

    class TestOrElseRaise:
        """or_else_raise method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_value(self) -> None:
                """When value is provided, return the value"""
                # GIVEN
                value: int = 1
                supplier: Supplier[Exception] = lambda: ValueError("Test error message")
                optional: JOptional[int] = JOptional.of_noneable(value)

                # WHEN
                result: int = optional.or_else_raise(supplier)

                # THEN
                assert result == value

            def test_none_value__should__raise_supplied_exception(self) -> None:
                """When no value is provided, raise the supplied exception"""
                # GIVEN
                given_error: Exception = ValueError("Test error message")
                supplier: Supplier[Exception] = lambda: given_error
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(ValueError) as error:
                    optional.or_else_raise(supplier)

                assert str(error.value) == str(given_error)

        class TestErrorCase:
            def test_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When no value and no supplier are provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_else_raise(None)

            def test_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When value is provided but supplier is none, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of_noneable(value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_else_raise(None)

    class TestOrGet:
        """or_get method tests"""

        class TestNominalCase:
            def test_value_exists__should__return_self(self) -> None:
                """When value is provided, return the self"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)
                supplier: Supplier[JOptional[int]] = lambda: JOptional.of(2)

                # WHEN
                result: JOptional[int] = optional.or_get(supplier)

                # THEN
                assert result == optional

            def test_none_value__should__return_supplied_joptional(self) -> None:
                """When value is not provided, return the supplied default JOptional"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                other: JOptional[int] = JOptional.of(2)
                supplier: Supplier[JOptional[int]] = lambda: other

                # WHEN
                result: JOptional[int] = optional.or_get(supplier)

                # THEN
                assert result == other

        class TestErrorCase:
            def test_value_exists_and_none_supplier__should__raise_type_error(self) -> None:
                """When value is provided but supplier is not, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_get(None)

            def test_none_value_and_none_supplier__should__raise_type_error(self) -> None:
                """When no value and no supplier are provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.or_get(None)

    class TestMap:
        """map method tests"""

        class TestNominalCase:
            def test_value_exists__should__joptional_of_mapped_value(self) -> None:
                """When value is provided, should apply the given mapper"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)
                mapper: Function[int, str] = lambda i: f"Str value is: '{i}'"

                # WHEN
                result: JOptional[str] = optional.map(mapper)

                # THEN
                assert result.is_present()
                assert result._value == "Str value is: '1'"

            def test_value_exists_and_none_mapper_result__should__empty(self) -> None:
                """When value is provided, should return an empty JOptional"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)
                mapper: Function[int, None] = lambda i: None

                # WHEN
                result: JOptional[str] = optional.map(mapper)

                # THEN
                assert result.is_empty()

            def test_none_value__should__return_empty(self) -> None:
                """When value is not provided, should return an empty JOptional"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                mapper: Function[int, str] = lambda i: f"Str value is: '{i}'"

                # WHEN
                result: JOptional[str] = optional.map(mapper)

                # THEN
                assert result.is_empty()

        class TestErrorCase:
            def test_value_exists_and_none_mapper__should__raise_type_error(self) -> None:
                """When value is provided but mapper is not, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.map(None)

            def test_none_value_and_none_mapper__should__raise_type_error(self) -> None:
                """When value and mapper are not provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                with raises(TypeError):
                    optional.map(None)

    class TestFlatMap:
        """flat_map method tests"""

        class TestNominalCase:
            def test_value_exists__should__joptional_of_mapped_value(self) -> None:
                """When value is provided, should apply the given mapper"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)
                mapper: Function[int, JOptional[str]] = \
                    lambda v: JOptional.of_noneable(v).map(lambda i: f"Str value is: '{i}'")

                # WHEN
                result: JOptional[str] = optional.flat_map(mapper)

                # THEN
                assert result.is_present()
                assert result._value == "Str value is: '1'"

            def test_none_value__should__return_empty(self) -> None:
                """When value is not provided, should return an empty JOptional"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                mapper: Function[int, JOptional[str]] = \
                    lambda v: JOptional.of_noneable(v).map(lambda i: f"Str value is: '{i}'")

                # WHEN
                result: JOptional[str] = optional.flat_map(mapper)

                # THEN
                assert result.is_empty()

        class TestErrorCase:
            def test_value_exists_and_none_mapper__should__raise_type_error(self) -> None:
                """When value is provided but mapper is not, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(None)

            def test_none_value_and_none_mapper__should__raise_type_error(self) -> None:
                """When value and mapper are not provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(None)

            def test_value_exists_and_none_mapper_result__should__raise_type_error(self) -> None:
                """When value exists and mapper returns None, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.flat_map(lambda v: None)

    class TestIfPresent:
        """if_present method tests"""

        class TestNominalCase:

            @patch('builtins.print')
            def test_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of(value)
                consumer: Consumer[[int]] = print

                # WHEN
                optional.if_present(consumer)

                # THEN
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            def test_none_value__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When no value is provided, should do nothing"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                consumer: Consumer[[int]] = print

                # WHEN
                optional.if_present(consumer)

                # THEN
                print_mock.assert_not_called()

        class TestErrorCase:
            def test_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is provided but consumer is not, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present(None)

            def test_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When no value and no consumer are provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present(None)

    class TestIfEmpty:
        """if_empty method tests"""

        class TestNominalCase:

            @patch('builtins.print')
            def test_value_exists__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When value is provided, should do nothing"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                optional.if_empty(runnable)

                # THEN
                print_mock.assert_not_called()

            @patch('builtins.print')
            def test_none_value__should__call_given_runnable(self, print_mock: MagicMock) -> None:
                """When no value is provided, should call the given runnable"""
                # GIVEN
                message: str = "A test message"
                optional: JOptional[int] = JOptional.empty()
                runnable: Runnable = lambda: print(message)

                # WHEN
                optional.if_empty(runnable)

                # THEN
                print_mock.assert_called_once_with(message)

        class TestErrorCase:
            def test_none_value_and_none_runnable__should__raise_type_error(self) -> None:
                """When no value and no consummer are provided, should raise TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_empty(cast(Runnable, None))

            def test_value_exists_and_none_runnable__should__raise_type_error(self) -> None:
                """When value is provided but runnable is not, should raise TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_empty(cast(Runnable, None))

    class TestIfPresentOrElse:
        """if_present_or_else method tests"""

        class TestNominalCase:
            @patch('builtins.print')
            def test_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of(value)
                consumer: Consumer[[int]] = print
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                optional.if_present_or_else(consumer, runnable)

                # THEN
                print_mock.assert_called_once_with(value)

            @patch('builtins.print')
            def test_none_value__should__call_given_runnable(self, print_mock: MagicMock) -> None:
                """When value is not provided, should call the given runnable"""
                # GIVEN
                message: str = "A test message"
                optional: JOptional[int] = JOptional.empty()
                consumer: Consumer[[int]] = print
                runnable: Runnable = lambda: print("A test message")

                # WHEN
                optional.if_present_or_else(consumer, runnable)

                # THEN
                print_mock.assert_called_once_with(message)

        class TestErrorCase:
            def test_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is provided but consumer is None, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of(value)
                runnable: Runnable = lambda: print("A test message")

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present_or_else(None, runnable)

            def test_value_exists_and_none_consumer_and_runnable__should__raise_type_error(self) -> None:
                """When value is provided but consumer and runnable are None, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of(value)

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present_or_else(None, cast(Runnable, None))

            def test_none_value_and_consumer_and_runnable__should__raise_type_error(self) -> None:
                """When value is not provided but consumer and runnable are None, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present_or_else(None, cast(Runnable, None))

            def test_none_value_runnable__should__raise_type_error(self) -> None:
                """When value is not provided and runnable is None, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                consumer: Consumer[[int]] = print

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present_or_else(consumer, cast(Runnable, None))

            def test_value_exists_and_none_runnable__should__raise_type_error(self) -> None:
                """When value is provided but not the runnable, should raise a TypeError exception"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of(value)
                consumer: Consumer[[int]] = print

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present_or_else(consumer, cast(Runnable, None))

            def test_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is not provided and consumer is None, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                runnable: Runnable = lambda: print("A test message")

                # WHEN / THEN
                with raises(TypeError):
                    optional.if_present_or_else(None, runnable)

    class TestFilter:
        """filter method tests"""

        class TestNominalCase:
            def test_value_with_valid_filter__should__return_joptional_with_value(self):
                """When value is present and predicate is valid, should return the JOptional with value"""
                # GIVEN
                value: int = 7
                predicate: Predicate[int] = lambda i: i > 5
                optional: JOptional[int] = JOptional.of(value)

                # WHEN
                result: JOptional[int] = optional.filter(predicate)

                # THEN
                assert result.is_present()
                assert result._value == value

            def test_value_without_valid_filter__should__return_empty(self):
                """When value is present and predicate is not valid, should return empty JOptional"""
                # GIVEN
                value: int = 7
                predicate: Predicate[int] = lambda i: i < 5
                optional: JOptional[int] = JOptional.of(value)

                # WHEN
                result: JOptional[int] = optional.filter(predicate)

                # THEN
                assert result.is_empty()

            def test_empty_value__should__return_empty(self):
                """When value is empty, should return empty JOptional"""
                # GIVEN
                predicate: Predicate[int] = lambda i: i < 5
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: JOptional[int] = optional.filter(predicate)

                # THEN
                assert result.is_empty()

        class TestErrorCase:
            def test_value_exists_and_predicate_is_none__should__raise_type_error(self) -> None:
                """When value is provided but predicate is None, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.filter(None)

            def test_empty_value_and_none_predicate__should__raise_type_erro(self):
                """When value is empty and predicate is None, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.filter(None)

    class TestPeek:
        """peek method tests"""

        class TestNominalCase:

            @patch('builtins.print')
            def test_value_exists__should__call_given_consumer(self, print_mock: MagicMock) -> None:
                """When value is provided, should call the given consumer"""
                # GIVEN
                value: int = 1
                optional: JOptional[int] = JOptional.of(value)
                consumer: Consumer[[int]] = print

                # WHEN
                result: JOptional[int] = optional.peek(consumer)

                # THEN
                print_mock.assert_called_once_with(value)
                assert result == optional

            @patch('builtins.print')
            def test_none_value__should__do_nothing(self, print_mock: MagicMock) -> None:
                """When no value is provided, should do nothing"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()
                consumer: Consumer[[int]] = print

                # WHEN
                result: JOptional[int] = optional.peek(consumer)

                # THEN
                print_mock.assert_not_called()
                assert result == optional

        class TestErrorCase:
            def test_value_exists_and_none_consumer__should__raise_type_error(self) -> None:
                """When value is provided but consumer is not, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN / THEN
                with raises(TypeError):
                    optional.peek(None)

            def test_none_value_and_none_consumer__should__raise_type_error(self) -> None:
                """When no value and no consumer are provided, should raise a TypeError exception"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN / THEN
                with raises(TypeError):
                    optional.peek(None)

    class TestIsAwaitable:
        """is_awaitable method tests"""

        class TestNominalCase:
            def test_none__should__return_false(self) -> None:
                """When value is None should return False"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: bool = optional.is_awaitable()

                # THEN
                assert not result

            async def test_coroutine_value__should__return_true(self) -> None:
                """When value is a Coroutine should return True"""

                # GIVEN
                async def coroutine() -> None: ...

                optional: JOptional[Awaitable[None]] = JOptional.of(coroutine())

                # WHEN
                result: bool = optional.is_awaitable()

                # THEN
                assert result
                await optional.get()  # Remove warning

            async def test_task_value__should__return_true(self) -> None:
                """When value is a Task should return True"""

                # GIVEN
                async def coroutine() -> None: ...

                optional: JOptional[Task] = JOptional.of(asyncio.create_task(coroutine()))

                # WHEN
                result: bool = optional.is_awaitable()

                # THEN
                assert result
                await optional.get()  # Remove warning

            async def test_future_value__should__return_true(self) -> None:
                """When value is a Future should return True"""

                # GIVEN
                async def coroutine() -> None: ...

                optional: JOptional[Future] = JOptional.of(asyncio.ensure_future(coroutine()))

                # WHEN
                result: bool = optional.is_awaitable()

                # THEN
                assert result
                await optional.get()  # Remove warning

            def test_not_awaitable_value__should__return_false(self) -> None:
                """When value is not an awaitable should return False"""
                # GIVEN
                optional: JOptional[str] = JOptional.of("A value")

                # WHEN
                result: bool = optional.is_awaitable()

                # THEN
                assert not result

    class TestToSyncValue:
        """to_sync_value method tests"""

        class TestNominalCase:
            async def test_none_value__should__return_self(self) -> None:
                """When value is None should return self"""
                # GIVEN
                optional: JOptional[int] = JOptional.empty()

                # WHEN
                result: JOptional[int] = await optional.to_sync_value()

                # THEN
                assert result == optional

            async def test_not_awaitable_value__should__return_self(self) -> None:
                """When value is not Awaitable should return self"""
                # GIVEN
                optional: JOptional[int] = JOptional.of(1)

                # WHEN
                result: JOptional[int] = await optional.to_sync_value()

                # THEN
                assert result == optional

            async def test_coroutine_value__should__return_joptional_with_synchronized_value(self) -> None:
                """When value is a Coroutine should return a JOptional describing the awaited value"""

                # GIVEN
                value: str = "A string value"

                async def coroutine() -> str: return value

                optional: JOptional[Awaitable[str]] = JOptional.of(coroutine())

                # WHEN
                result: JOptional[str] = await optional.to_sync_value()

                # THEN
                assert result.get() == value

            async def test_task_value__should__return_joptional_with_synchronized_value(self) -> None:
                """When value is a Task should return a JOptional describing the awaited value"""

                # GIVEN
                value: str = "A string value"

                async def coroutine() -> str: return value

                optional: JOptional[Awaitable[str]] = JOptional.of(asyncio.create_task(coroutine()))

                # WHEN
                result: JOptional[str] = await optional.to_sync_value()

                # THEN
                assert result.get() == value

            async def test_future_value__should__return_joptional_with_synchronized_value(self) -> None:
                """When value is a Future should return a JOptional describing the awaited value"""

                # GIVEN
                value: str = "A string value"

                async def coroutine() -> str: return value

                optional: JOptional[Awaitable[str]] = JOptional.of(asyncio.ensure_future(coroutine()))

                # WHEN
                result: JOptional[str] = await optional.to_sync_value()

                # THEN
                assert result.get() == value

    class TestToAsyncJOptional:
        """to_async_joptional method tests"""

        class TestNominalCase:
            async def test_empty__should__return_empty_async_joptional(self) -> None:
                """When value is None should return empty AsyncJOptional"""
                # GIVEN
                optional: JOptional[str] = JOptional.empty()

                # WHEN
                async_optional: AsyncJOptional[str] = optional.to_async_joptional()

                # THEN
                assert await async_optional.is_empty()

            async def test_str_value__should__return_str_async_joptional(self) -> None:
                """When str value should return an AsyncJOptional[str] with value"""
                # GIVEN
                value: str = "A test value"
                optional: JOptional[str] = JOptional.of(value)

                # WHEN
                async_optional: AsyncJOptional[str] = optional.to_async_joptional()

                # THEN
                assert await async_optional.is_present()
                assert await async_optional.get() == value

            async def test_async_str_value__should__return_str_async_joptional(self) -> None:
                """When str value should return an AsyncJOptional[str] with value"""
                # GIVEN
                value: str = "A test value"
                optional: JOptional[Awaitable[str]] = JOptional.of(to_awaitable(value))

                # WHEN
                async_optional: AsyncJOptional[str] = optional.to_async_joptional()

                # THEN
                assert await async_optional.is_present()
                assert await async_optional.get() == value
